/*
	local
	2004, Andreas Romeyke
	mailto:art1@it-netservice.de

*/

/** \mainpage
\section intro Introduction
\subsection motiv Motivation
The complete package was developed by Andreas Romeyke in case of his master-thesis.
\subsection eatr EATR -- EEG ARTIFACTS TRAINER
The file eatr.c (compiled "eatr") is the main file to train an ANN (arteficial neural network) to detect artifacts, e.g. eye-blinks in EEG-data. A network-file will be created, to load it with ear.c to remove artifacts automagically.

\subsection ear EAR -- EEG ARTIFACTS REJECTER
The file ear.c (compiled "ear") is the main file to load a trained ANN to detect artefacts automagically. It is feeded with the decompressed EEG-data and creates an REJ-file.

\subsection cntconv CNTCONV -- a cnt desqueezer
The file cntconv.c (compiled "cntconv") is the main file to extract compressed EEG-data (in cnt-fileformat) to an uncompressed binary or human-readable stream. The human-readable will be in GnuPlot compatible format, the binary will only be used within all the programs of the EATR/EAR-project. The EEG-data will also be normalized and prepared for easier access. This program was separated, because licensing and political reasons.

\section LICENSE
All the stuff, except cntconv.c and acntconv.c licensed under the terms of the GNU GENERAL PUBLIC LICENSE (GPL), please see http://fsf.org for details. There should also be a file with a complete copy of the GPL license into this package.

The files cntconv.c and acntconv.c were not free available until the opencntlib will be licensed under the terms of the GNU General Public LICENSE. In this case they will also be free available under the terms of GPL.
*/

/* TODO: calc global sample-idx, random preselect stuff from filelist in net.infiles */
/* TODO: change EEGBUF-Fileformat to add a value for "norm" data in cntconvert.c, change to write in filename instead
stdout, join acntconv */
/* DONE: add version support in all files */
/* DONE: implement in eatr a test-network to calc the MQGE, prepare splitted data, add config support for this */

#ifndef _MY_LOCAL_H
#define _MY_LOCAL_H

/** time() for srand()-init */
#include <time.h>

/** fopen and Co. */
#include <stdio.h>

/** printf and Co. */
#include <stdlib.h>

/** error number */
#include <errno.h>

/** strerror(), strcmp(), strcat() etc. */
#include <string.h>

/** sqrt() */
#include <math.h>

/** set assertions, "gcc -DNDEBUG" disables assertions */
#include <assert.h>

/** test if a float/double is a NaN-value (not a number) */
#define ISNAN(a) (((a)!=(a)))
/* a NaN value compared with itself will be failed ever */

/** The value used in cntcnv.c to normalize EEG-data to avaoid neuron-locking, because the actual implementation of libfann uses a tanh or sigmoid function with an saturation of +-20, means, that the value will be constant +-1 if these limits will be reached. */
#define NORMVALUE 400
/* 65536 */

/** The actual sample frequency of all used EEG-data is here 250 Hz */
#define SAMPLEFREQUENCY 250
/* 250 Hz */

/** all programs will be compiled with this count of required EEG-data-channels. \warning only files created with same value of required EEG-channels will be interact sanely */
#define REQUIREDCHANNELS 6
/* 62 */

/** this defines the actual used windowsize. This size should be a multiple of DOWNSAMPLE. 128 or 256 were a good choice at 250 Hz, this value depends on SAMPLEFREQUENCY. */
#define MAXTAPS 256
/* 256 */
/* 200 */ /* was 100 for start15 */

/** this is needed for downsampling of data, remember MAXTAPS should be an multiple of this. */
#define DOWNSAMPLE 8

/** this is the count of different statistical features in convert_stat(), it is 10, because there will be the range, average, standard deviation, skewness and kurtosis calculated for two different windowsizes. */
#define STATFEATURES 10

/** TRUE definition that way to be portable */
#define TRUE (1==1)
/** FALSE definition that way to be portable */
#define FALSE (1==0)


/** originally used by eatr.c and ear.c to report which conversion was used */
enum e_convert {
	dummy=1, /** 1:1 copy as input for ANN */
	downsample=2,
	summarize_electrodes=4,
	sfft=8,
	stat=16 /** using statistical descriptions as input for ANN */
};

/** originally used by cntconv to report which normalization was used */
enum e_norm {
	simple=1, /** only a simple scaling with a constant factor */
	minmaxwindow,
	minmaxenvelope
};

/** count of entries in e_channelmap without dummy */
#define MAX_MAPPED_CHANNELS 23
/** enumeration to map into the International 10-20-system */
enum e_channelmap {
	eogh=1,
	eogv,
	fp1,
	fp2,
	f7,
	f8,
	f3,
	fz,
	f4,
	t3,
	c3,
	cz,
	c4,
	t4,
	a1,
	a2,
	p3,
	pz,
	p4,
	t5,
	t6,
	o1,
	o2,
	empty
};


/** this stores necessary values in decompressed EEG-data */
struct s_cntconvhead {
	/** count of channels */
	unsigned int channelcount;
	/** count of samples per channel (all channels with same count) */
	unsigned int samplecount;
	/** which normalization is used */
	enum e_norm norm;
};

/** this structure stores internally a set of corresponding data  */
struct s_eegfile {
	/** decompressed EEG datafile */
	char * eeg_filename;
	/** correspondent datafile for reject-data (REJ) */
	char * rej_filename;
	/** correspondent datafile for channel identification, not used yet */
	char * chn_filename;
	/** holds the readed EEG-data for buffered access, already converted to global index */
	float * eeg_buffer;
	/* holds the readed REJ-data for buffered access, already converted to global index */
	unsigned short int * rej_buffer;
	unsigned short int channelcount;
	/** max sample count of this file */
	unsigned int samplecount;
};

/** this defines return-values if the programs will be failed, useful only if scripting is needed */
enum earerr{
	/** all okay */
	ERROK=0,
	/** the cmdline argument is wrong or unexpected */
	ERRARG,
	/** there were errors in parsing streams/files */
	ERRSTREAM,
	/** unexpected channel count */
	ERRCHANCOUNT,
	/** error in memory allocation */
	ERRMEM,
	/** unexpected error in program code */
	ERRPRGCODE,
	/** error in libfann */
	ERRFANN,
	/** error in trasnforming data */
	ERRCONVERTDATA,
	/** the given input files differ in channelcount or electrode configuration */
	ERRDIFFFILES
};

/* prototypes */


/* cntconv.c */
int parse_cnt(char * filename, unsigned short int * channelcount, unsigned int * samplecount, float ** ppcbuf);
int print_chanlabels(char * filename);
int write_header(unsigned short int * channelcount, unsigned int * samplecount, unsigned short int outputformat);
int write_data(unsigned short int * channelcount, unsigned int * samplecount, float ** pcbuff, unsigned short int outputformat);
/* shared/parseeeg.c */
int parse_eeg(char * filename, unsigned short int * chanc, unsigned int * samplec, float ** ppebuf);
void free_eeg(float ** ppebuf);
/* shared/parserej.c */
void free_rej(unsigned short int ** pprbuf);
int parse_rej(char * filename, unsigned int * samplecount, unsigned short int ** prbuff);
/* shared/mrandomize.c */
unsigned int mrandomize(unsigned int max);

#endif
