/** \file
	this file contains a various set of convert/transformation functions to extract data for a learning step of the ANN.
*/
#include <rfftw.h>
#include <fftw.h>
#include "local_ann.h"

/** the input data will be downsampled by factor DOWNSAMPLE. Over a window of size DOWNSAMPLE the average value will be calculated and returned. */
unsigned long int convert_downsample(unsigned long int countofelements, fann_type ** ppin, fann_type ** ppout) {
	fann_type * pout;
	fann_type * pin;
	unsigned short int channel;
	unsigned short int taps;
	unsigned short int subsample;
	assert(NULL != ppin);
	assert(NULL != ppout);
	assert(countofelements >0);
	pout= *ppout;
	pin= *ppin;
	assert(NULL != pin);
	assert(NULL == pout);
	assert((MAXTAPS%DOWNSAMPLE)==0); /* secure to avoid rounding problems */
	pout=malloc((size_t) (countofelements * sizeof(fann_type *))/DOWNSAMPLE);
	if (NULL==pout) {
		fprintf(stderr, "could not allocate memory for buffer, %s\n", strerror(errno));
		exit(ERRMEM);
	}
	assert(DOWNSAMPLE>0);
	for (channel=0; channel<MAX_MAPPED_CHANNELS; channel++) {
		for (taps=0; taps<MAXTAPS; taps=taps+(unsigned short int) DOWNSAMPLE) {
			assert(taps<MAXTAPS);
			pout[((taps*MAX_MAPPED_CHANNELS)/DOWNSAMPLE)+channel]=0.0f;
			for (subsample=0; subsample<DOWNSAMPLE; subsample++) {
				pout[((taps*MAX_MAPPED_CHANNELS)/DOWNSAMPLE)+channel]+= pin[(taps+subsample)*MAX_MAPPED_CHANNELS+channel];
			}
			pout[((taps*MAX_MAPPED_CHANNELS)/DOWNSAMPLE)+channel]/=(float) DOWNSAMPLE;
		}
	}
	*ppout=pout;
	return((MAXTAPS*MAX_MAPPED_CHANNELS)/DOWNSAMPLE);
}

#ifdef NOTDEFINED
					tmpinput[taps*net.channelcount+l]=0;
					for (m=1; m<DOWNSAMPLE; m++) {
						tmpinput[taps*net.channelcount+l]+= (fann_type)
						net.input[(k+taps+m)*net.channelcount+l];
					}
					tmpinput[taps*net.channelcount+l]=tmpinput[taps*net.channelcount+l]/DOWNSAMPLE;
#endif

/** Over all channels the sample-points at a time will be added to a new value */
unsigned long int convert_summarize_electrodes(unsigned long int countofelements, fann_type ** ppin, fann_type ** ppout){
	fann_type * pout;
	fann_type * pin;
	unsigned short int channel;
	unsigned short int taps;

	assert(NULL != ppin);
	assert(NULL != ppout);
	assert(countofelements >0);
	pout= *ppout;
	pin= *ppin;
	assert(NULL != pin);
	assert(NULL == pout);
	pout=malloc((size_t) countofelements * sizeof(fann_type *)/MAX_MAPPED_CHANNELS );
	if (NULL==pout) {
		fprintf(stderr, "could not allocate memory for buffer, %s\n", strerror(errno));
		exit(ERRMEM);
	}
	for (taps=0; taps<MAXTAPS; taps++) {
		pout[taps]=0.0f;
		for (channel=0; channel<MAX_MAPPED_CHANNELS; channel++) {
			pout[taps]=pout[taps]+pin[taps*MAX_MAPPED_CHANNELS+channel];
		}
	}
	*ppout=pout;
	return(MAXTAPS);
}

#define WITH_SFFT 1
#ifdef WITH_SFFT
/** does a Gabor-transformation of a specified window */
unsigned long int convert_shorttime_fft(unsigned long int countofelements, fann_type ** ppin, fann_type ** ppout){
	fftw_real tmpfftin[MAXTAPS];
	fftw_real tmpfftout[MAXTAPS];
	rfftw_plan plan;
 	int l;
	fann_type * pout;
	fann_type * pin;
	unsigned short int taps;
	assert(NULL != ppin);
	assert(NULL != ppout);
	assert(countofelements >0);
	pout= *ppout;
	pin= *ppin;
	assert(NULL != pin);
	assert(NULL == pout);
	pout=malloc((size_t) countofelements * sizeof(fann_type *) );
	if (NULL==pout) {
		fprintf(stderr, "could not allocate memory for buffer, %s\n", strerror(errno));
		exit(ERRMEM);
	}
			/* eventually transform data with fft */
			/* shorttime-FFT here */
			plan= rfftw_create_plan(MAXTAPS, FFTW_FORWARD, FFTW_ESTIMATE);
			for (l=0; l<MAX_MAPPED_CHANNELS; l++) { 					assert(l<MAX_MAPPED_CHANNELS);
				for (taps=0; taps<MAXTAPS; taps++) {
					assert(taps<MAXTAPS);
					tmpfftin[taps]=pin[taps*MAX_MAPPED_CHANNELS+l];
				}
				rfftw_one(plan, tmpfftin, tmpfftout);
				for (taps=0; taps<MAXTAPS; taps++) {
					assert(taps<MAXTAPS);
					pout[taps*MAX_MAPPED_CHANNELS+l]=tmpfftout[taps];
				}
			}
			fftw_destroy_plan(plan);
	*ppout=pout;
	return(countofelements);
}
#endif

/** this does nothing than pass-through the raw EEG-data */
unsigned long int convert_dummy(unsigned long int countofelements, fann_type ** ppin, fann_type ** ppout) {
	fann_type * pout;
	fann_type * pin;
	unsigned short int channel;
	unsigned short int taps;
	fann_type temporary_storage;
	assert(NULL != ppin);
	assert(NULL != ppout);
	assert(countofelements >0);
	pout= *ppout;
	pin= *ppin;
	assert(NULL != pin);
	assert(NULL == pout);
	pout=malloc((size_t) countofelements * sizeof(fann_type *) );
	if (NULL==pout) {
		fprintf(stderr, "could not allocate memory for buffer, %s\n", strerror(errno));
		exit(ERRMEM);
	}
	/* memcpy(pout, pin, (size_t) countofelements * sizeof(fann_type *)); */
	for (channel=0; channel<MAX_MAPPED_CHANNELS; channel++) {
		for (taps=0; taps<MAXTAPS; taps++) {
			temporary_storage=	pin[(taps*MAX_MAPPED_CHANNELS)+channel];
			pout[(taps*MAX_MAPPED_CHANNELS)+channel]=temporary_storage;
		}
	}
	*ppout=pout;
	return(countofelements);
}

/** The raw EEG-data will be passed through, an additional channel will be created with the sum of samplepoints over all channels at a time. There was the hope to detect electrode-artifacts with the ANN. */
unsigned long int convert_default_plus_sum(unsigned long int countofelements, fann_type ** ppin, fann_type ** ppout) {
	fann_type * pout;
	fann_type * pin;
	unsigned short int taps;
	unsigned short int channel;
	float sum;
	assert(NULL != ppin);
	assert(NULL != ppout);
	assert(countofelements >0);
	pout= *ppout;
	pin= *ppin;
	assert(NULL != pin);
	assert(NULL == pout);
	pout=malloc((size_t) (countofelements+MAXTAPS )* sizeof(fann_type *) );
	if (NULL==pout) {
		fprintf(stderr, "could not allocate memory for buffer, %s\n", strerror(errno));
		exit(ERRMEM);
	}
	for (taps=0;taps<MAXTAPS; taps++) {
		sum=0;
		for (channel=0; channel<MAX_MAPPED_CHANNELS; channel++) {
			pout[(taps*MAX_MAPPED_CHANNELS+1)+channel]=pin[(taps*MAX_MAPPED_CHANNELS)+channel];
			sum=sum+pin[(taps*MAX_MAPPED_CHANNELS)+channel];
		}
		pout[(taps*MAX_MAPPED_CHANNELS+1)+channel+1]=sum;
	}
	*ppout=pout;
	return(countofelements+MAXTAPS);
}


/** transforms per channel a set of samplepoints (of size MAXTAPS) to various statistical values. Actually there were 10 statistical features (as defined with STATFEATURES):
 - Average of half window
 - Standard Deviation of halfwindow
 - Range of halfwindow
 - Skewness of half window \f$ \frac{\sum_{i=0}^{N}(x_{i}-\mu)^3}{Ns^3}\f$
 - Kurtosis of half window \f$ \frac{\sum_{i=0}^{N}(x_{i}-\mu)^4}{Ns^4}-3\f$
 - all above with full sized window
*/
unsigned long int convert_stat(unsigned long int countofelements, fann_type ** ppin, fann_type ** ppout) {
	fann_type * pout;
	fann_type * pin;
	unsigned short int channel;
	unsigned short int taps;
	float sum;
	float dsqrsum;
	float tsqrsum;
	float qsqrsum;
	float diffmean;
	float stddev;
	float min;
	float max;
	assert(NULL != ppin);
	assert(NULL != ppout);
	assert(countofelements >0);
	pout= *ppout;
	pin= *ppin;
	assert(NULL != pin);
	assert(NULL == pout);
	pout=malloc((size_t) MAX_MAPPED_CHANNELS * STATFEATURES * sizeof(fann_type *) );
	if (NULL==pout) {
		fprintf(stderr, "could not allocate memory for buffer, %s\n", strerror(errno));
		exit(ERRMEM);
	}
	for (channel=0; channel<MAX_MAPPED_CHANNELS; channel++) {
		pout[(0*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* 1/2 avg */
		pout[(1*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* 1/2 stdev */
		pout[(2*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* avg */
		pout[(3*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* stdv */
		pout[(4*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* 1/2 range */
		pout[(5*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* range */
		pout[(6*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* 1/2 skewness */
		pout[(7*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* 1/2 kurtosis */
		pout[(8*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* skewness */
		pout[(9*MAX_MAPPED_CHANNELS)+channel]=0.0f; /* kurtosis */
		/* pout[(6*MAX_MAPPED_CHANNELS)+channel]=0; */ /* minmax-freq */
		sum=0; min=10000; max=-10000;
		for (taps=0; taps<MAXTAPS/2; taps++) {
			sum+=pin[(taps*MAX_MAPPED_CHANNELS)+channel];
			if (min>pin[(taps*MAX_MAPPED_CHANNELS)+channel])
				min=pin[(taps*MAX_MAPPED_CHANNELS)+channel];
			if (max<pin[(taps*MAX_MAPPED_CHANNELS)+channel])
				max=pin[(taps*MAX_MAPPED_CHANNELS)+channel];
		}
		pout[(4*MAX_MAPPED_CHANNELS)+channel]=max-min; /* 1/2 range */
		pout[(0*MAX_MAPPED_CHANNELS)+channel]=sum * (1/(float) (MAXTAPS/2)); /* 1/2 avg */
		for (taps=MAXTAPS/2; taps<MAXTAPS; taps++) {
			sum+=pin[(taps*MAX_MAPPED_CHANNELS)+channel];
			if (min>pin[(taps*MAX_MAPPED_CHANNELS)+channel])
				min=pin[(taps*MAX_MAPPED_CHANNELS)+channel];
			if (max<pin[(taps*MAX_MAPPED_CHANNELS)+channel])
				max=pin[(taps*MAX_MAPPED_CHANNELS)+channel];
		}
		pout[(5*MAX_MAPPED_CHANNELS)+channel]=max-min; /* range */
		pout[(2*MAX_MAPPED_CHANNELS)+channel]=sum/(float) MAXTAPS; /* avg */
		dsqrsum=0;
		tsqrsum=0;
		qsqrsum=0;
		for (taps=0; taps<MAXTAPS/2; taps++) {
			diffmean=(pin[(taps*MAX_MAPPED_CHANNELS)+channel]-pout[(0*MAX_MAPPED_CHANNELS)+channel]);
			dsqrsum+=(diffmean * diffmean);
			tsqrsum+=(diffmean * diffmean * diffmean);
			qsqrsum+=(diffmean * diffmean * diffmean * diffmean);
		}
		stddev=(float)sqrt(dsqrsum/(((double)MAXTAPS/2)-1)); /* 1/2 stdev */
		if (stddev!=0.0f) {
			pout[(1*MAX_MAPPED_CHANNELS)+channel]=stddev;
			pout[(6*MAX_MAPPED_CHANNELS)+channel]=tsqrsum* (1/((float)(MAXTAPS/2)*stddev*stddev*stddev)); /* 1/2 skewness */
/*			printf("tsqrsum: %f, stddev: %f, stddev^3: %f, all: %f\n",  tsqrsum, stddev,stddev*stddev*stddev,  tsqrsum* (1/((float)(MAXTAPS/2)*stddev*stddev*stddev))); */
			pout[(7*MAX_MAPPED_CHANNELS)+channel]=qsqrsum/((float)(MAXTAPS/2)*stddev*stddev*stddev*stddev)-3; /* 1/2 kurtosis */
		}
		dsqrsum=0;
		tsqrsum=0;
		qsqrsum=0;
		stddev=0;
		for (taps=0; taps<MAXTAPS; taps++) {
			diffmean=(pin[(taps*MAX_MAPPED_CHANNELS)+channel]-pout[(2*MAX_MAPPED_CHANNELS)+channel]);
			dsqrsum+=(diffmean * diffmean);
			tsqrsum+=(diffmean * diffmean * diffmean);
			qsqrsum+=(diffmean * diffmean * diffmean * diffmean);
		}
		stddev=(float)sqrt(dsqrsum/(double)(MAXTAPS-1)); /* stdv */
		if (stddev!=0.0f) {
			pout[(3*MAX_MAPPED_CHANNELS)+channel]=stddev;
			pout[(8*MAX_MAPPED_CHANNELS)+channel]=(float)(tsqrsum/((double)(MAXTAPS)*stddev*stddev*stddev)); /* skewness */
			pout[(9*MAX_MAPPED_CHANNELS)+channel]=(float)(qsqrsum/((double)(MAXTAPS)*stddev*stddev*stddev*stddev))-3; /*  kurtosis */
			/*printf("NAN: channel %i, value: %f isnan: %i\n", channel, pout[(6*MAX_MAPPED_CHANNELS)+channel]), isnan(pout[(6*MAX_MAPPED_CHANNELS)+channel]); */
		}
		assert(ISNAN(pout[(0*MAX_MAPPED_CHANNELS)+channel])==0);
		assert(ISNAN(pout[(1*MAX_MAPPED_CHANNELS)+channel])==0);
		assert(ISNAN(pout[(2*MAX_MAPPED_CHANNELS)+channel])==0);
		assert(ISNAN(pout[(3*MAX_MAPPED_CHANNELS)+channel])==0);
		assert(ISNAN(pout[(4*MAX_MAPPED_CHANNELS)+channel])==0);
		assert(ISNAN(pout[(5*MAX_MAPPED_CHANNELS)+channel])==0);
		assert(ISNAN(pout[(6*MAX_MAPPED_CHANNELS)+channel])==0);
		assert(ISNAN(pout[(7*MAX_MAPPED_CHANNELS)+channel])==0);
		assert(ISNAN(pout[(8*MAX_MAPPED_CHANNELS)+channel])==0);
		assert(ISNAN(pout[(9*MAX_MAPPED_CHANNELS)+channel])==0);
	}
	*ppout=pout;
	return(MAX_MAPPED_CHANNELS*STATFEATURES);
}

/** wrapper function to encapsulate the different convert-functions */
unsigned long int convert(unsigned long int countofelements, fann_type ** ppin, fann_type ** ppout) {
#ifdef NOTDEFINED
	return(convert_dummy(countofelements, ppin, ppout));
#else
	switch (net.convert) {
		case dummy: return(convert_dummy(countofelements, ppin, ppout)); break;
		case downsample: return(convert_downsample(countofelements, ppin, ppout)); break;
		case summarize_electrodes: return(convert_summarize_electrodes(countofelements, ppin, ppout)); break;
		case sfft: return(convert_shorttime_fft(countofelements, ppin, ppout)); break;
		case stat: return(convert_stat(countofelements, ppin, ppout)); break;
		default: fprintf(stderr, "wrong parameter for \"-c %i\" (wrong enum_value) or error in code line: %i, file: %s\n", net.convert, __LINE__, __FILE__);
		exit(ERRARG);
	}
#endif
	return(0);
}


void free_convert(fann_type ** ppout) {
	fann_type * pout;
	assert(NULL != ppout);
	pout= *ppout;
	assert(NULL != pout);
	free(pout);
	pout=NULL;
	*ppout=pout;
}

